{% extends "base.html" %}

{% block title %}All Movies{% endblock %}

{% block content %}
<h2>All Movies</h2>
<ul>
  {% for movie in movies %}
  <li><a href="{{ url_for('movie', id=movie.id) }}">{{ movie.title }}</a></li>
  {% endfor %}
</ul>
{% endblock %}
