<!DOCTYPE html>
<html>
  <head>
    <title>{% block title %}{% endblock %} - Movie Database</title>
    <link rel="stylesheet" type="text/css"
          href="{{ url_for('static', path='style.css') }}">
  </head>
  <body>
    <div id="title">
      <a href="{{ url_for('/') }}"><h1>Movie Database</h1></a>
    </div>
    <div id="content">
{% block content %}{% endblock %}
    </div>
  </body>
</html>
